/*
q3shaders.h - macro replacement that converts Quake3 shaders into XashNT materials
Copyright (C) 2019 Uncle Mike

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
*/

// shaders case are nonsense
#caseinsensative

#include "scripts\q3const.h"

enum stageNum;
enum texModCount;
enum deformCount;
enum numBlends;
enum numAdditiveBlends;
enum stageHasDepthWrite;
enum transMaterial;

// radiant keywords (gone them)
#keydef qer_editorimage <value>
#keydef qer_alphafunc <a> <b>
#keydef qer_trans <a>
#define qer_nocarve

// q3map keywords (just eliminate)
#define q3map_noFog
#define q3map_globalTexture
#define q3map_nonplanar
#keydef q3map_backShader <path>
#keydef q3map_shadeAngle <a>
#keydef q3map_sun <a> <b> <c> <d> <e> <f>
#keydef q3map_sunExt <a> <b> <c> <d> <e> <f> <g> <h>
#keydef q3map_tcGen <a> <b> <c> <d> <e> <f> <g> <h> <j> <k> <i>
#keydef q3map_alphaMod <a> <b> <c> <d> <e> <f>
#keydef q3map_skyLight <a> <b>
#keydef q3map_lightImage <a>
#keydef q3map_surfacelight <a>
#keydef q3map_bounceScale <a>
#keydef q3map_cloneShader <a>
#keydef q3map_vertexScale <a>
#define q3map_invert
#keydef q3map_alphaMod <a>
#keydef q3map_alphaMod <set> <a>
#keydef q3map_alphaMod <func> <lbracket> <a> <b> <c> <rbracket>
#keydef q3map_lightsubdivide <a>
#keydef q3map_backSplash <a> <b>
#define q3map_forceMeta
#define q3map_notjunc
#keydef q3map_bounce <a>
#keydef q3map_fur <a> <b> <c>
#keydef q3map_flare <shaderName>
#keydef q3map_lightmapFilterRadius <a> <b>
#keydef q3map_normalImage <path>
#keydef q3map_textureSize <x> <y>
#define q3map_lightmapMergable
#keydef q3map_lightmapAxis <axis>
#keydef q3map_tcMod <type> <val>
#keydef q3map_lightmapSampleSize <ssize>
#keydef q3map_lightRGB <r> <g> <b>
#define q3map_clipModel

/*
#keydef #beginsection\
	#msg <sectionname> started
*/

#keydef #endsection\
	CheckBlend

// surfaceparms, etc
#keydef surfaceparm <value>

// ugly method to hide clip surfaces
#keydef surfaceparm clip\
	alphaFunc( GL_NEVER, 0.0f );

#keydef surfaceparm botclip\
	alphaFunc( GL_NEVER, 0.0f );

#keydef surfaceparm nodraw\
	alphaFunc( GL_NEVER, 0.0f );

// FIXME: handle skyparms too
#keydef skyparms <a> <b> <c>\
	addShaderDefine( SKYPARMS );

#keydef tessSize <a>
#define nomipmaps
#keydef depthFunc <a>\
//	#if <stageNum> == 0\
//		addShaderDefine( SCREEN_COLOR );\
//	#endif\
	addShaderDefine( DEPTH_EQUAL<stageNum> );

#define nopicmip
#define detail
#keydef light <a>
#define fogonly
#define light1
#keydef sort <a>

// NOTE: fog volumes coulnd't correctly blending with standard methods, sigh
#keydef fogparms <a> <b> <c> <d> <e>\
	addShaderDefine( SCREEN_COLOR );\
	setFogParams( <a>, <b>, <c>, <d> );

#keydef fogparms ( <a> <b> <c> ) <d>\
	addShaderDefine( SCREEN_COLOR );\
	setFogParams( <a>, <b>, <c>, <d> );

#keydef sky <path>
#keydef cloudparms <a>
#keydef cloudparms <a> <b>
#define lightning
#define entityMergable

#keydef CountBlend\
	#if u_BlendFunc0[0] == 1.0 && u_BlendFunc0[1] == 1.0\
		numAdditiveBlends++;\
	#endif\
	#if u_BlendFunc<stageNum>[0] != 1.0 || u_BlendFunc<stageNum>[1] != 0.0\
		#if <stageNum> == 0\
			transMaterial = 1;\
		#endif\
		numBlends++;\
	#endif

#keydef CheckHardwareBlend\
	#if u_BlendFunc0[0] == 1.0 && u_BlendFunc0[1] == 1.0\
		addShaderDefine( TRANSLUCENT );\
		blendFunc( GL_ONE, GL_ONE );\
		depthMask( GL_FALSE );\
	#elif u_BlendFunc0[0] == 3.0 && u_BlendFunc0[1] == 1.0\
		addShaderDefine( TRANSLUCENT );\
		blendFunc( GL_DST_COLOR, GL_ONE );\
		depthMask( GL_FALSE );\
	#elif u_BlendFunc0[0] == 3.0 && u_BlendFunc0[1] == 0.0\
		addShaderDefine( TRANSLUCENT );\
		blendFunc( GL_DST_COLOR, GL_ZERO );\
		depthMask( GL_FALSE );\
	#elif u_BlendFunc0[0] == 1.0 && u_BlendFunc0[1] == 2.0\
		addShaderDefine( TRANSLUCENT );\
		blendFunc( GL_ONE, GL_SRC_COLOR );\
		depthMask( GL_FALSE );\
	#elif u_BlendFunc0[0] == 6.0 && u_BlendFunc0[1] == 1.0\
		addShaderDefine( TRANSLUCENT );\
		blendFunc( GL_SRC_ALPHA, GL_ONE );\
		depthMask( GL_FALSE );\
	#elif u_BlendFunc0[0] == 6.0 && u_BlendFunc0[1] == 8.0\
		addShaderDefine( TRANSLUCENT );\
		blendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );\
		depthMask( GL_FALSE );\
	#elif u_BlendFunc0[0] == 0.0 && u_BlendFunc0[1] == 4.0\
		addShaderDefine( TRANSLUCENT );\
		blendFunc( GL_ZERO, GL_ONE_MINUS_SRC_COLOR );\
		depthMask( GL_FALSE );\
	#elif u_BlendFunc0[0] == 0.0 && u_BlendFunc0[1] == 8.0\
		addShaderDefine( TRANSLUCENT );\
		blendFunc( GL_ZERO, GL_ONE_MINUS_SRC_ALPHA );\
		depthMask( GL_FALSE );\
	#endif

#keydef CheckBlend\
	#ifndef SCREEN_COLOR\
		#if <transMaterial> == 1\
			#if <numBlends> == 1 || <numBlends> == <numAdditiveBlends>\
				CheckHardwareBlend\
			#else\
				#ifndef SKYPARMS\
					#msg <sectionname> apply custom blend\
					addShaderDefine( SCREEN_COLOR );\
					depthMask( GL_TRUE );\
				#endif\
			#endif\
		#endif\
	#endif

#keydef CheckAlpha\
	#if <stageNum> == 0\
	#if u_AlphaFunc0[0] == 2.0\
		alphaFunc( GL_GREATER, 0.0f );\
	#endif\
	#if u_AlphaFunc0[0] == 3.0\
		alphaFunc( GL_LESS, 0.5f );\
	#endif\
	#if u_AlphaFunc0[0] == 4.0\
		alphaFunc( GL_GEQUAL, 0.5f );\
	#endif\
	#endif

#keydef CheckCustomBlend\
	#ifndef SKYPARMS\
		#if <stageHasDepthWrite> && <stageNum> != 0\
			#msg <sectionname> apply custom blend\
			addShaderDefine( SCREEN_COLOR );\
			depthMask( GL_TRUE );\
		#endif\
		// reset variable
		stageHasDepthWrite = 0;\
	#endif

// a method to determine transparent surfaces
#keydef surfaceparm trans

// if shader has depthWrite we should use custom blend
#keydef depthWrite\
	stageHasDepthWrite = 1;

// replace key 'portal' with implicit stage (should be first)
#keydef portal\
	addShaderDefine( PLANAR_REFLECTION );\
	addShaderDefine( SHADER_STAGE<stageNum> );\
// FIXME: replace with subview when multi-pass will be finalized
//	image u_ColorMap<stageNum> = "entity->$SubviewTexture";\
	image u_ColorMap<stageNum> = "textures\common\mirror1.tga";\
	vec2 u_BlendFunc<stageNum> = vec2( 1.0, 0.0 );\
	addShaderDefine( TEXGEN_UNIT<stageNum> var_TexMirror );\
	texModCount = 0;\
	stageNum++;

#keydef polygonOffset\
	polygonOffset( -1.0f, -2.0f );

#keydef map <value>\
	image u_ColorMap<stageNum> = "<value>";

#keydef animMap <fps> <frame0>\
	image u_ColorMap<stageNum> = "<frame0>";

#keydef animMap <fps> <frame0> <frame1>\
	image u_ColorMap<stageNum> = "<frame0>";\
	addUnitFrame( u_ColorMap<stageNum>, <frame1>, 1 );\
	setUnitFramerate( u_ColorMap<stageNum>, <fps> );

#keydef animMap <fps> <frame0> <frame1> <frame2>\
	image u_ColorMap<stageNum> = "<frame0>";\
	addUnitFrame( u_ColorMap<stageNum>, <frame1>, 1 );\
	addUnitFrame( u_ColorMap<stageNum>, <frame2>, 2 );\
	setUnitFramerate( u_ColorMap<stageNum>, <fps> );

#keydef animMap <fps> <frame0> <frame1> <frame2> <frame3>\
	image u_ColorMap<stageNum> = "<frame0>";\
	addUnitFrame( u_ColorMap<stageNum>, <frame1>, 1 );\
	addUnitFrame( u_ColorMap<stageNum>, <frame2>, 2 );\
	addUnitFrame( u_ColorMap<stageNum>, <frame3>, 3 );\
	setUnitFramerate( u_ColorMap<stageNum>, <fps> );

#keydef animMap <fps> <frame0> <frame1> <frame2> <frame3> <frame4>\
	image u_ColorMap<stageNum> = "<frame0>";\
	addUnitFrame( u_ColorMap<stageNum>, <frame1>, 1 );\
	addUnitFrame( u_ColorMap<stageNum>, <frame2>, 2 );\
	addUnitFrame( u_ColorMap<stageNum>, <frame3>, 3 );\
	addUnitFrame( u_ColorMap<stageNum>, <frame4>, 4 );\
	setUnitFramerate( u_ColorMap<stageNum>, <fps> );

#keydef animMap <fps> <frame0> <frame1> <frame2> <frame3> <frame4> <frame5>\
	image u_ColorMap<stageNum> = "<frame0>";\
	addUnitFrame( u_ColorMap<stageNum>, <frame1>, 1 );\
	addUnitFrame( u_ColorMap<stageNum>, <frame2>, 2 );\
	addUnitFrame( u_ColorMap<stageNum>, <frame3>, 3 );\
	addUnitFrame( u_ColorMap<stageNum>, <frame4>, 4 );\
	addUnitFrame( u_ColorMap<stageNum>, <frame5>, 5 );\
	setUnitFramerate( u_ColorMap<stageNum>, <fps> );

#keydef animMap <fps> <frame0> <frame1> <frame2> <frame3> <frame4> <frame5> <frame6>\
	image u_ColorMap<stageNum> = "<frame0>";\
	addUnitFrame( u_ColorMap<stageNum>, <frame1>, 1 );\
	addUnitFrame( u_ColorMap<stageNum>, <frame2>, 2 );\
	addUnitFrame( u_ColorMap<stageNum>, <frame3>, 3 );\
	addUnitFrame( u_ColorMap<stageNum>, <frame4>, 4 );\
	addUnitFrame( u_ColorMap<stageNum>, <frame5>, 5 );\
	addUnitFrame( u_ColorMap<stageNum>, <frame6>, 6 );\
	setUnitFramerate( u_ColorMap<stageNum>, <fps> );

#keydef animMap <fps> <frame0> <frame1> <frame2> <frame3> <frame4> <frame5> <frame6> <frame7>\
	image u_ColorMap<stageNum> = "<frame0>";\
	addUnitFrame( u_ColorMap<stageNum>, <frame1>, 1 );\
	addUnitFrame( u_ColorMap<stageNum>, <frame2>, 2 );\
	addUnitFrame( u_ColorMap<stageNum>, <frame3>, 3 );\
	addUnitFrame( u_ColorMap<stageNum>, <frame4>, 4 );\
	addUnitFrame( u_ColorMap<stageNum>, <frame5>, 5 );\
	addUnitFrame( u_ColorMap<stageNum>, <frame6>, 6 );\
	addUnitFrame( u_ColorMap<stageNum>, <frame7>, 7 );\
	setUnitFramerate( u_ColorMap<stageNum>, <fps> );

#keydef clampMap <value>\
	image u_ColorMap<stageNum> = "<value>";\
	addImageFlags( u_ColorMap<stageNum>, TF_CLAMP );
	
#keydef map $lightmap\
	addShaderDefine( TEXGEN_UNIT<stageNum> var_TexLightMap );\
	image u_ColorMap<stageNum> = "entity->$LightmapTexture";

// blend presets
#keydef blendFunc add\
	vec2 u_BlendFunc<stageNum> = vec2( 1.0, 1.0 );\
	CountBlend

#keydef blendFunc blend\
	vec2 u_BlendFunc<stageNum> = vec2( 6.0, 8.0 );\
	CountBlend

#keydef blendFunc filter\
	vec2 u_BlendFunc<stageNum> = vec2( 3.0, 0.0 );\
	CountBlend
	
#keydef blendFunc <src> <dst>\
	vec2 u_BlendFunc<stageNum> = vec2( <src>, <dst> );\
	CountBlend

// check for first unit only
#keydef alphaFunc GT0\
	vec2 u_AlphaFunc<stageNum> = vec2( 2.0, 0.0 );\
	CheckAlpha

#keydef alphaFunc LT128\
	vec2 u_AlphaFunc<stageNum> = vec2( 3.0, 0.5 );\
	CheckAlpha

#keydef alphaFunc GE128\
	vec2 u_AlphaFunc<stageNum> = vec2( 4.0, 0.5 );\
	CheckAlpha

// rgbGen's
#keydef rgbGen wave <func> <base> <amp> <phase> <freq>\
	addShaderDefine( RGBGEN_UNIT<stageNum> ColorWave( <func>, vec4( <base>, <amp>, <phase>, <freq> )) );

#keydef rgbGen const ( <a> <b> <c> )\
	addShaderDefine( RGBGEN_UNIT<stageNum> vec3( <a>, <b>, <c> ) );

// just skip it
#keydef rgbGen identityLighting
#keydef rgbGen identity

#keydef rgbGen entity\
	addShaderDefine( RGBGEN_UNIT<stageNum> u_RenderColor.rgb );

#keydef rgbGen oneMinusEntity\
	addShaderDefine( RGBGEN_UNIT<stageNum> (1.0 - u_RenderColor.rgb) );

// NOTE: i hope is correct
#keydef rgbGen vertex\
	addShaderDefine( RGBGEN_UNIT<stageNum> var_VertexColor.rgb );\
	// it's will be applied if alphaGen was not specified by user
	addShaderDefine( IMPLICIT_ALPHA<stageNum> );

#keydef rgbGen exactVertex\
	addShaderDefine( RGBGEN_UNIT<stageNum> var_VertexColor.rgb );

#keydef rgbGen oneMinusVertex\
	addShaderDefine( RGBGEN_UNIT<stageNum> (1.0-var_VertexColor.rgb) );

// FIXME: implement when lightgrid will be finalized in the engine
#keydef rgbGen lightingdiffuse

// alphaGen's
#keydef alphaGen wave <func> <base> <amp> <phase> <freq>\
	addShaderDefine( ALPHAGEN_UNIT<stageNum> AlphaWave( <func>, vec4( <base>, <amp>, <phase>, <freq> )) );

#keydef alphaGen const <a>\
	addShaderDefine( ALPHAGEN_UNIT<stageNum> ( <a> ) );

// just skip it
#keydef alphaGen identity

#keydef alphaGen entity\
	addShaderDefine( ALPHAGEN_UNIT<stageNum> u_RenderColor.a );

#keydef alphaGen oneMinusEntity\
	addShaderDefine( ALPHAGEN_UNIT<stageNum> (1.0 - u_RenderColor.a) );

#keydef alphaGen vertex\
	addShaderDefine( ALPHAGEN_UNIT<stageNum> var_VertexColor.a );

#keydef alphaGen oneMinusVertex\
	addShaderDefine( ALPHAGEN_UNIT<stageNum> (1.0 - var_VertexColor.a) );

#keydef alphaGen lightingSpecular\
	addShaderDefine( ALPHAGEN_UNIT<stageNum> CalcSpecular() );

#keydef alphaGen portal <a>\
	addShaderDefine( ALPHAGEN_UNIT<stageNum> CalcFade( <a> ) );

// tex mods
#keydef tcMod rotate <speed>\
	addShaderDefine( APPLY_TEXMOD<stageNum>_<texModCount> );\
	vec4 u_texModBase<stageNum>_<texModCount> = vec4( <rotate>, <speed>, 0.0f, 0.0f );\
	vec4 u_texModFunc<stageNum>_<texModCount> = vec4( 0.0f, 0.0f, 0.0f, 0.0f );\
	texModCount++;

#keydef tcMod rotate <speed> <unused>\
	addShaderDefine( APPLY_TEXMOD<stageNum>_<texModCount> );\
	vec4 u_texModBase<stageNum>_<texModCount> = vec4( <rotate>, <speed>, 0.0f, 0.0f );\
	vec4 u_texModFunc<stageNum>_<texModCount> = vec4( 0.0f, 0.0f, 0.0f, 0.0f );\
	texModCount++;

#keydef tcMod scroll <x> <y>\
	addShaderDefine( APPLY_TEXMOD<stageNum>_<texModCount> );\
	vec4 u_texModBase<stageNum>_<texModCount> = vec4( <scroll>, <x>, <y>, 0.0f );\
	vec4 u_texModFunc<stageNum>_<texModCount> = vec4( 0.0f, 0.0f, 0.0f, 0.0f );\
	texModCount++;

// old weird format
#keydef tcMod scroll <x> <y> <unused1> <unused2>\
	addShaderDefine( APPLY_TEXMOD<stageNum>_<texModCount> );\
	vec4 u_texModBase<stageNum>_<texModCount> = vec4( <scroll>, <x>, <y>, 0.0f );\
	vec4 u_texModFunc<stageNum>_<texModCount> = vec4( 0.0f, 0.0f, 0.0f, 0.0f );\
	texModCount++;

#keydef tcMod scale <x> <y>\
	addShaderDefine( APPLY_TEXMOD<stageNum>_<texModCount> );\
	vec4 u_texModBase<stageNum>_<texModCount> = vec4( <scale>, <x>, <y>, 0.0f );\
	vec4 u_texModFunc<stageNum>_<texModCount> = vec4( 0.0f, 0.0f, 0.0f, 0.0f );\
	texModCount++;

// old weird format
#keydef tcMod scale <x> <y> <unused1> <unused2>\
	addShaderDefine( APPLY_TEXMOD<stageNum>_<texModCount> );\
	vec4 u_texModBase<stageNum>_<texModCount> = vec4( <scale>, <x>, <y>, 0.0f );\
	vec4 u_texModFunc<stageNum>_<texModCount> = vec4( 0.0f, 0.0f, 0.0f, 0.0f );\
	texModCount++;

#keydef tcMod stretch <func> <base> <amp> <phase> <freq>\
	addShaderDefine( APPLY_TEXMOD<stageNum>_<texModCount> );\
	vec4 u_texModBase<stageNum>_<texModCount> = vec4( <stretch>, <func>, 0.0f, 0.0f );\
	vec4 u_texModFunc<stageNum>_<texModCount> = vec4( <base>, <amp>, <phase>, <freq> );\
	texModCount++;

#keydef tcMod turb <base> <amp> <phase> <freq>\
	addShaderDefine( APPLY_TEXMOD<stageNum>_<texModCount> );\
	vec4 u_texModBase<stageNum>_<texModCount> = vec4( <turb>, 0.0f, 0.0f, 0.0f );\
	vec4 u_texModFunc<stageNum>_<texModCount> = vec4( <base>, <amp>, <phase>, <freq> );\
	texModCount++;

// old method. <func> is not used
#keydef tcMod turb <func> <base> <amp> <phase> <freq>\
	addShaderDefine( APPLY_TEXMOD<stageNum>_<texModCount> );\
	vec4 u_texModBase<stageNum>_<texModCount> = vec4( <turb>, 0.0f, 0.0f, 0.0f );\
	vec4 u_texModFunc<stageNum>_<texModCount> = vec4( <base>, <amp>, <phase>, <freq> );\
	texModCount++;

#keydef tcMod transform <a> <b> <c> <d> <e> <f>\
	addShaderDefine( APPLY_TEXMOD<stageNum>_<texModCount> );\
	vec4 u_texModBase<stageNum>_<texModCount> = vec4( <a>, <b>, <c>, 1.0f );\
	vec4 u_texModFunc<stageNum>_<texModCount> = vec4( <d>, <e>, <f>, 1.0f );\
	texModCount++;

#keydef tcGen base\
	addShaderDefine( TEXGEN_UNIT<stageNum> var_TexDiffuse );

#keydef tcGen texture\
	addShaderDefine( TEXGEN_UNIT<stageNum> var_TexDiffuse );

// NOTE: this will not work, but is never used ;-)
#keydef tcGen lightmap\
	addShaderDefine( TEXGEN_UNIT<stageNum> var_TexLightMap );

#keydef tcGen environment\
	addShaderDefine( COMPUTE_ENVMAP );\
	addShaderDefine( TEXGEN_UNIT<stageNum> var_TexEnvMap );

#keydef tcGen vector ( <x0> <y0> <z0> ) ( <x1> <y1> <z1> )\
	addShaderDefine( TEXGEN_UNIT<stageNum> var_TexAutoGen );\
	vec3 u_texGenS = vec3( <x0>, <y0>, <z0> );\
	vec3 u_texGenT = vec3( <x1>, <y1>, <z1> );\
	addShaderDefine( GENERATE_AUTOCOORDS );

// vertex deforms
#keydef deformVertexes projectionShadow
#keydef deformVertexes text0
#keydef deformVertexes text1

#keydef deformVertexes autosprite\
	setGeometryHint( AUTOSPRITE );\
	addShaderDefine( APPLY_DEFORM<deformCount> );\
	vec2 u_DeformType<deformCount> = vec2( <autosprite>, 0.0 );\
	vec4 u_DeformBase<deformCount> = vec4( 0.0, 0.0, 0.0, 0.0 );\
	vec3 u_DeformMisc<deformCount> = vec3( 0.0, 0.0, 0.0 );\
	deformCount++;

#keydef deformVertexes autosprite2\
	setGeometryHint( AUTOSPRITE2 );\
	// some data goes into tangent and binormal heh
	addShaderDefine( COMPUTE_TBN );\
	addShaderDefine( APPLY_DEFORM<deformCount> );\
	vec2 u_DeformType<deformCount> = vec2( <autosprite2>, 0.0 );\
	vec4 u_DeformBase<deformCount> = vec4( 0.0, 0.0, 0.0, 0.0 );\
	vec3 u_DeformMisc<deformCount> = vec3( 0.0, 0.0, 0.0 );\
	deformCount++;

#keydef deformVertexes wave <spread> <waveform> <base> <amp> <phase> <freq>\
	addShaderDefine( APPLY_DEFORM<deformCount> );\
	vec2 u_DeformType<deformCount> = vec2( <wave>, <waveform> );\
	vec4 u_DeformBase<deformCount> = vec4( <base>, <amp>, <phase>, <freq> );\
	vec3 u_DeformMisc<deformCount> = vec3( <spread>, 0.0, 0.0 );\
	deformCount++;

#keydef deformVertexes move <x> <y> <z> <waveform> <base> <amp> <phase> <freq>\
	addShaderDefine( APPLY_DEFORM<deformCount> );\
	vec2 u_DeformType<deformCount> = vec2( <move>, <waveform> );\
	vec4 u_DeformBase<deformCount> = vec4( <base>, <amp>, <phase>, <freq> );\
	vec3 u_DeformMisc<deformCount> = vec3( <x>, <y>, <z> );\
	deformCount++;

#keydef deformVertexes normal <freq> <amp>\
	addShaderDefine( APPLY_DEFORM<deformCount> );\
	vec2 u_DeformType<deformCount> = vec2( <normal>, <noise> );\
	vec4 u_DeformBase<deformCount> = vec4( 0.0, <amp>, 0.0, <freq> );\
	vec3 u_DeformMisc<deformCount> = vec3( 0.0, 0.0, 0.0 );\
	deformCount++;

#keydef deformVertexes bulge <width> <height> <speed>\
	addShaderDefine( APPLY_DEFORM<deformCount> );\
	vec2 u_DeformType<deformCount> = vec2( <bulge>, <sin> );\
	vec4 u_DeformBase<deformCount> = vec4( 0.0, 0.0, 0.0, 0.0 );\
	vec3 u_DeformMisc<deformCount> = vec3( <width>, <height>, <speed> );\
	deformCount++;

#keydef {\
	addShaderDefine( SHADER_STAGE<stageNum> );

#keydef }\
	// add default blendFunc if it was not specified early
	vec2 u_BlendFunc<stageNum> = vec2( 1.0, 0.0 );\
	// add default alphaFunc if it was not specified early
	vec2 u_AlphaFunc<stageNum> = vec2( 1.0, 0.0 );\
	// add default texGen if it was not specified early
	addShaderDefine( TEXGEN_UNIT<stageNum> var_TexDiffuse );\
	CheckCustomBlend\
	texModCount = 0;\
	stageNum++;

#keydef cull none\
	cullFace( GL_NONE );
	
#keydef cull disable\
	cullFace( GL_NONE );

#keydef cull twosided\
	cullFace( GL_NONE );

#keydef cull front\
	cullFace( GL_FRONT );

#keydef cull back\
	cullFace( GL_BACK );